/*
Program: watchd
Created: 20.08.2001
Author : Peter Turczak <p_turczak@wiwa.de>
Syntax : watchd

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define __USE_BSD

#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>

// Adapt here, if this it NOT CHANGED CORRECTLY, overflows may be a result :-(
// Just specify the amount of folders to be watched at maximum in one session!
#define MAXFOLDERS 10

typedef struct Twatchfolder 
		{
		 char* runprg;
		 char* dir;
		} Twatchfolder;

Twatchfolder folders[MAXFOLDERS]; 
int usedir=0;   // Number of used folders, must ALWAYS be <=MAXFOLDERS!

int interval=60;// Rescan interval in seconds
char *tmpdir;   // Directory where the processing dirs are created. Default=/tmp
char *mvprg;    // Program that is used to move the files around. Default=mv 
char *rmprg;    // Program that is used to remove the files. Default=rm 

int nolock(char *fn)
{
int fd=open(fn,O_RDWR);
             /* l_type   l_whence  l_start  l_len  l_pid   */
struct flock fl = { F_WRLCK, SEEK_SET, 0,       0,     0 };

fl.l_pid = getpid();
fl.l_type = F_WRLCK;

if (fcntl(fd, F_SETLK, &fl) == -1) 
 {
 #ifdef DEBUG
  printf("nolock(%s)=%d\n",fn,0);
 #endif 
 return(0);  
 } else {
 #ifdef DEBUG
  printf("nolock(%s)=%d\n",fn,1);
 #endif 
 fl.l_type = F_UNLCK;  /* set to unlock same region */
 if (fcntl(fd, F_SETLK, &fl) == -1) 
  {
  #ifdef DEBUG
  printf("Strange unlock behavior..\n");
  #endif
  }
  return(1);
 }
    

}

int splits(char *s, char* p1, char* p2)
{
 char *equal=strchr(s,'=');
 char *p2tmp;
 int pos1 = (((equal==NULL)||(equal<s))?0:(equal-s)); 

 s[pos1]=0x00;			// Needed, because strncpy is too stupid to 
 				// null-terminate a substring by itself!
 strcpy(p1,"");
 strcpy(p2,"");

 if (pos1==0) return(1); else 
 {
 strncpy(p1, s, pos1+1);
 strcpy(p2, equal+1);

// p2tmp=malloc(strlen(p2));
// strncpy(p2tmp,p2,strlen(p2)-1);
// strcpy(p2,p2tmp);
// free(p2tmp);

 p2tmp=strchr(p2,0xa);		// Not very nice,
 p2[(int)p2tmp-(int)p2]=0x00;	// but in fact: It works ;)

 while (p2[0]==' ') {p2++;}       		     // Again, quick and dirty..
 while (p1[strlen(p1)-1]==' ') {p1[strlen(p1)-1]=0;} // Need to say anything?;)

 #ifdef DEBUG
 printf("split(s='%s',p1='%s',p2='%s',pos1=%d);\n",s,p1,p2,pos1); 
 #endif
 return(0);
 }
 
}

void addfolder(int nr, char *dir, char *prg)
{
 if (usedir<nr) {
  usedir=nr;
  folders[usedir-1].runprg=malloc(1024);
  folders[usedir-1].dir=malloc(1024);
  strcpy(folders[usedir-1].runprg,"");  // We are clean, aren't we...
  strcpy(folders[usedir-1].dir,"");	// dito..
  } else 
  {
   #ifdef DEBUG
   printf("Entry %d already allocated, no need to reallocate it!\n",nr); 
   #endif 
  }
 if (prg!=NULL) strcpy(folders[usedir-1].runprg,prg);  
 if (dir!=NULL) strcpy(folders[usedir-1].dir,dir);  
}

void lowercase(char *c)
{
int i;

for (i=0;i<strlen(c);i++)
 {
  c[i]=tolower(c[i]);
 }
}

void parseequal(char* buf) 
{
 char *tmpc1;
 char *tmpc2;
 char *tmpc3;
 int tmpi;   // a place to store temorary integers ...

 tmpc1=malloc(strlen(buf)+1);
 tmpc2=malloc(strlen(buf)+1);

 splits(buf, tmpc1, tmpc2);
 
 lowercase(tmpc1);      // Drop the case on the left side...
 
 #ifdef DEBUG
 printf("lowercase returned '%s'!\n",tmpc1);
 #endif 
 
 if (strncmp(tmpc1,"prg",3)==0) 
  {
   #ifdef DEBUG
   printf("'prg' line found. Adding program...");
   #endif
   tmpc3=tmpc1+3;
   tmpi=atoi(tmpc3);
   #ifdef DEBUG
   printf("using nr. %d (%s) \n",tmpi,tmpc3);
   #endif
   addfolder(tmpi,NULL,tmpc2);
  }

 if (strncmp(tmpc1,"dir",3)==0) 
  {
   #ifdef DEBUG
   printf("'dir' line found. Adding directory...");
   #endif
   tmpc3=tmpc1+3;
   tmpi=atoi(tmpc3);
   #ifdef DEBUG
   printf("using nr. %d (%s) \n",tmpi,tmpc3);
   #endif
   addfolder(tmpi,tmpc2,NULL);
  }
 
 if (strncmp(tmpc1,"interval",8)==0) 
  {
   #ifdef DEBUG
   printf("'interval' line found. Trying to set interval...");
   #endif
   
   tmpc3=tmpc2;
   tmpi=atoi(tmpc3);
   if (tmpi>0) {interval=tmpi;}// set interval if the value is not completely
   			       // brain-damaged!
   #ifdef DEBUG
   printf("set to %d!\n",interval);
   #endif
  }
 
 if (strncmp(tmpc1,"tempdir",7)==0) 
  {
   #ifdef DEBUG
   printf("'tempdir' line found. Trying to set tempdir...");
   #endif
   strcpy(tmpdir,tmpc2);
   if (tmpdir[strlen(tmpdir)-1]!='/') {strcat(tmpdir,"/");}
   #ifdef DEBUG
   printf("set to %s!\n",tmpdir);
   #endif
  }

 if (strncmp(tmpc1,"mv",7)==0) 
  {
   #ifdef DEBUG
   printf("'mv' line found. Trying to set mvprg...");
   #endif
   strcpy(mvprg,tmpc2);
   #ifdef DEBUG
   printf("set to %s!\n",mvprg);
   #endif
  }
   
 if (strncmp(tmpc1,"rm",7)==0) 
  {
   #ifdef DEBUG
   printf("'rm' line found. Trying to set mvprg...");
   #endif
   strcpy(rmprg,tmpc2);
   #ifdef DEBUG
   printf("set to %s!\n",rmprg);
   #endif
  }
   
 free(tmpc1);
 free(tmpc2);
}

void defraglst()
{
int i;
int i2;
 
for (i=0;i<usedir;i++) 
 {
  if ((strcmp(folders[i].dir,"")==0)|(strcmp(folders[i].runprg,"")==0))
   {
    printf("Entry #%d not completely defined, both dir and prg line must be present! Ignoring this entry...\n",i+1);
    free(folders[i].dir);
    free(folders[i].runprg);
    usedir--;	
    for (i2=i;i2<usedir;i2++) 
    {
     folders[i2]=folders[i2+1];
    } 
   }
 }
}

void init()
{
 FILE *f;
 char *buf;
 int i;

 tmpdir=malloc(1024); 
 strcpy(tmpdir,"/tmp/");
 
 mvprg=malloc(1024); 
 strcpy(mvprg,"/bin/mv");
 
 rmprg=malloc(1024); 
 strcpy(rmprg,"/bin/rm");
 
 buf=malloc(1024);
// buf=malloc(1024);
// buf=malloc(1024);
 f=fopen("/etc/watchd.conf","r");
 if (!f) 
  {
   printf("Error: Could not open /etc/watchd.conf, exiting..\n");
   exit(1);
  }
 while ((!feof(f)) && (usedir<MAXFOLDERS))
  {
   strcpy(buf,"");            // Clean up, before working..
   fgets(buf,1024,f);
   if ((buf[0]!='#') && (strlen(buf)>0))
    {
     if (strstr(buf,"=")!=NULL) {
      parseequal(buf); 
     } else {
     usedir++;
     folders[usedir-1].runprg=malloc(1024);
     folders[usedir-1].dir=malloc(1024);
     sscanf(buf,"%s%s",folders[usedir-1].dir,folders[usedir-1].runprg); 
     syslog(LOG_INFO||LOG_DAEMON,"Using old config style for entry #%d.",usedir-1);
     }
    } 
  }
 free(buf);
 fclose(f);
 defraglst();
 #ifdef DEBUG
 for (i=0;i<usedir;i++) { printf("Will look for files in '%s' and will run '%s' for them!\n",folders[i].dir,folders[i].runprg); }
 #endif
}

void run(char* cmd, char* arg1, char* arg2, char* arg3, char* arg4)
{
 char **args;
 int f;
 
 #ifdef DEBUG
 printf("run(cmd=%s,arg1=%s,arg2=%s,arg3=%s,arg4=%s)\n",cmd, arg1, arg2, arg3, arg4);
 #endif
 f=fork();
 if (f==0) {
  args=malloc(sizeof(args)*5); // allocate ram for 4 arguments...
  args[0]=cmd;
  args[1]=arg1;
  args[2]=arg2;
  args[3]=arg3;
  args[4]=arg4;
 #ifdef DEBUG
  printf("forked of...");
 #endif
  execvp(cmd,args);
  printf("Something went wrong! Could not execute %s...",cmd);
  syslog(LOG_WARNING||LOG_DAEMON,"Could not execute %s",cmd);
  exit(0); // Should never occur, but we are paranoid...
 } else 
 {
 #ifdef DEBUG
  printf("pid of child=%d\n",f);
 #endif
  waitpid(f,NULL,0); // wait for process to finish..
 #ifdef DEBUG
  printf("child died: pid=%d\n",f);
 #endif
  
 }
}

void processfile(char *file, char *srcdir, char *cmd)
{
char *complete;
char *srcfile;
struct timeb tp;

complete=malloc(255);
srcfile=malloc(255);

strcpy(srcfile,srcdir);
strcat(srcfile,"/");
strcat(srcfile,file);

if (nolock(srcfile)) {
 syslog(LOG_NOTICE||LOG_DAEMON,"processing %s with %s",file,cmd);
 
 ftime(&tp);
 sprintf(complete, "%s/watchdtmp.%d%d", tmpdir, tp.time, tp.millitm);
 
 mkdir(complete,0700);
 #ifdef DEBUG
 printf("created tempdir : %s\n",complete);
 #endif
 
#ifdef DEBUG
 printf("runing mv...\n");
 #endif
 run(mvprg,srcfile,complete,NULL,NULL);
 
 strcpy(srcfile,complete);
 strcat(srcfile,"/");
 
 strcat(srcfile,file);
 
 run(cmd,srcfile,NULL,NULL,NULL);
 run(rmprg,"-rf",complete,NULL,NULL);
} else {
 syslog(LOG_NOTICE||LOG_DAEMON,"file %s is locked, not processed",file,cmd);
}
free(complete);
free(srcfile);
//printf("flushing...\n");
fflush(stdout);
fflush(stdin);
fflush(stderr);
#ifdef DEBUG
printf("dying...\n");
#endif
exit(0);                 // Now the child dies..
#ifdef DEBUG
printf("damn, still there...\n");
#endif
}

int isdir(char *name)
{
 struct stat *st;

 st=malloc(sizeof(stat));
 stat(name, st);
 if (S_ISDIR(st->st_mode)) { 
   free(st);
   return(1); 
  } else { 
   free(st);
   return(0); 
   }
}

void processdir(int n)
{
 DIR* dir;
 struct dirent *de;
 char* tmpS;
 int child;

 if (folders[n].dir!=NULL) dir=opendir(folders[n].dir); else dir=NULL;
 tmpS=malloc(1024);
 if (dir!=NULL) 
  {
   de=readdir(dir);
   while (de!=NULL)
   { 
    if ((strcmp(de->d_name,".")!=0) && (strcmp(de->d_name,"..")!=0))  { // we don't process dirs, yet...
      child=fork();    
      #ifdef DEBUG
      if (child!=0) printf("child started, pid=%d\n",child);
      #endif
     if (child==0) // so, i am a child process...
      {
       #ifndef DEBUG
       fclose(stdin);
       fclose(stdout);
       #endif
       if (isdir(de->d_name)==0) {
         processfile(de->d_name,folders[n].dir,folders[n].runprg); 
	} else
	{
         syslog(LOG_WARNING||LOG_DAEMON,"sorry, cannot read dirs, currently ",folders[n].dir);
	 //processdir(de->d_name);
	}
      } else {
      #ifdef DEBUG 
        printf("waiting for child to die...(pid=%d)\n",child); 
	#endif
  	waitpid(child,NULL,0); // wait for process to finish..
        #ifdef DEBUG 
        printf("child died!!!(pid=%d)\n",child); 
	#endif
	}
      
    } 
     de=readdir(dir);
   }
  } else 
  { printf("Warning: Could not read directory '%s'\n",folders[n].dir); 
    syslog(LOG_WARNING||LOG_DAEMON,"could not read directory %s",folders[n].dir);
  }
  free(tmpS);
  closedir(dir);
}

void sighandler(int signal)
{
 if (signal==SIGHUP) 
  {
   init();  // Reread config file! 
  }
}

int main(int argc, char *argv[] )
{
 int i;
 
// signal(SIGHUP, sighandler);
 init();
 
 #ifndef DEBUG
 if (fork()==0) {
 #endif
 while (1) 
 { 
  for (i=0;i<usedir;i++)
   { 
    processdir(i);
   }
  sleep(interval);
 }
 #ifndef DEBUG
 }
 #endif 
}
